import json


class MockResponse:
    def __init__(self, status_code, text):
        self.status_code = status_code
        self.text = text or 'default_text'
        self.url = 'mock_url'

    def json(self):
        return json.loads(self.text)

    def text(self):
        return self.text


def read_file(path):
    with open(path) as f:
        return f.read()



"""Test user instances"""
from pytest import mark

from fawrapper.models import User


@mark.vcr
def test_pull_details(request_maker):
    real_uuid = "fd55615331be40bcb2758e9e585cff54"
    user = User(request_maker, uuid=real_uuid)
    assert user.first_name is None
    assert user.archived is None
    assert user.role is None
    assert user.last_name is None
    user.pull_details()
    assert user.archived is False
    assert user.role == 'Admin'
    assert user.last_name == 'Martinez'
    assert user.first_name == 'Chase'




#!/usr/bin/env python3
"""
Models of all resources on the api.
"""
from datetime import datetime
from fawrapper.models.base import ListResource, InstanceResource
from fawrapper.utils import format_time_to_iso

class Change(InstanceResource):

    def _set_attributes(self, params):
        self.status = params.get('status')
        self.changed_by = params.get('changedBy')
        self.entity_class = params.get('entityClass')
        self.url = params.get('url')
        self.entity = params.get('entity')
        self.actions = params.get('actions')
        self.id = params.get('id')


class Changes(ListResource):
    """
    Class for changes instance resource on the api
    Gets changes after a token, which is saved to disk
    for use on the next request.
    Token will be saved to the users home directory.
    """


class Task(InstanceResource):
    """
    Class for Task instance resource on the api
    glAccount mandatory name uuid markupType taxable
    markupValue customFields estDuration unitPrice unitCost
    description
    """
    path = '/task/'

    def _set_attributes(self, params):
        self.uuid = params.get('uuid')
        self.gl_account = params.get('glAccount')
        self.name = params.get('name')
        self.markup_type = params.get('markupType')
        self.taxable = params.get('taxable')
        self.markup_value = params.get('markupValue')
        self.custom_fields = params.get('customFields')
        self.est_duration = params.get('estDuration')
        self.unit_price = params.get('unitPrice')
        self.unit_cost = params.get('unitCost')
        self.description = params.get('description')


class Tasks(ListResource):
    instance = Task


class ItemInTask(InstanceResource):
    """Class for item in job object"""

    def _set_attributes(self, params):
        self.unit_cost = params.get('unitCost')
        self.unit_price = params.get('unitPrice')
        self.uuid = params.get('uuid')
        self.quantity = params.get('quantity')
        self.custom_fields = params.get('customFields')
        self._item = params.get('item')

    @property
    def item(self):
        item = Item(self.requester, **self._item)
        item.pull_details()
        return item


class TaskInJob(InstanceResource):
    """Class for task in job object"""

    def __init__(self, requester,  **params):
        self.job_uuid = params['job_uuid']
        self.path = f'/job/{self.job_uuid}/task/'
        super().__init__(requester, **params)

    def _set_attributes(self, params):
        self.attachments = params.get('attachments')
        self.done = params.get('done')
        self.unit_price = params.get('unitPrice')
        self.uuid = params.get('uuid')
        self.note = params.get('note')
        self.asset = params.get('asset')
        self.custom_fields = params.get('customFields')
        self.unit_cost = params.get('unitCost')
        self._task = params.get('task')
        self._items = params.get('items')

    @property
    def items(self):
        items_in_task_list = list()
        for item_dict in self._items:
            item_in_task = ItemInTask(self.requester, **item_dict)
            items_in_task_list.append(item_in_task)
        return items_in_task_list

    @property
    def task(self):
        task = Task(self.requester, **self._task)
        task.pull_details()
        return task


class Job(InstanceResource):
    """
    Class for Job instance resource on the api.

    Attributes from API:
    tasks description revenue signature jobId invoice jobLead
    scheduledOn labor pre_signature customer estDuration uuid
    createdOn crew state contact asset completedOn customFields
    location

    """
    path = '/job/'

    def _set_attributes(self, params):
        self.uuid = params.get('uuid')
        self.scheduled_on = params.get('scheduledOn')
        self.description = params.get('description')
        self.location = params.get('location')
        self.contact = params.get('contact')
        self.est_duration = params.get('estDuration')
        self.asset = params.get('asset')
        self.crew = params.get('crew')
        self.custom_fields = params.get('customFields')
        self.labor = params.get('labor')
        self.started_on = params.get('startedOn')
        self.revenue = params.get('revenue')
        self.signature = params.get('signature')
        self.job_id = params.get('jobId')
        self.invoice = params.get('invoice')
        self.pre_signature = params.get('pre_signature')
        self.customer = params.get('customer')
        self.created_on = params.get('createdOn')
        self.completed_on = params.get('completedOn')
        self._state = params.get('state')
        self._job_lead = params.get('jobLead')
        self._tasks = params.get('tasks')
        self._items_used = list()

    @property
    def items_used(self):
        for task in self.tasks:
            for item in task.items:
                self._items_used.append(item.item)
        return self._items_used

    @property
    def state(self):
        return self._state[0]

    @property
    def job_lead(self):
        if isinstance(self._job_lead, User):
            return self._job_lead
        else:
            user = User(self.requester, **self._job_lead)
            user.pull_details()
            self._job_lead = user
            return self._job_lead

    @job_lead.setter
    def job_lead(self, user):
        self._job_lead = user
        data = {"jobLead": {"uuid": user.uuid}}
        self.post(data)

    @property
    def tasks(self):
        tasks_list = list()
        for task_dict in self._tasks:
            task_dict.update({'job_uuid': self.uuid})
            task_instance = TaskInJob(self.requester, **task_dict)
            task_instance.pull_details()
            tasks_list.append(task_instance)
        return tasks_list


class Jobs(ListResource):
    """
    Class for Jobs list resource on the api
    """
    instance = Job
    valid_filters = ['start', 'end', 'sortedBy',
                     'location', 'jobLead', 'state']

    def list(self,filters=None):
        """
        Listing jobs will require a 'start' and 'end' value
        to be included in the filters.

        Query Parameters:
            start – starting date (required)
            end – end date (required)
            sortedBy – Available sorting criteria:
                scheduledOn, jobLead , location and state
            location – filtered by location
            jobLead – filtered by jobLead
            state – filtered by state

        """
        if not filters:
            return super().list()
        if 'pageSize' not in filters:
            filters['pageSize'] = 200
        filters = self._validate_filters(filters)
        return super().list(filters=filters)

    def _validate_filters(self, filters=None):
        filters = super()._validate_filters(filters)
        if 'start' not in filters:
            filters['start'] = format_time_to_iso(2000, 1, 1)
        if 'end' not in filters:
            end = datetime(2060, 1, 1)
            filters['end'] = format_time_to_iso(
                end.year, end.month, end.day, end.hour, end.minute)
        return filters


class User(InstanceResource):
    """
    Class for User instance resource
    """
    path = '/user/'

    def _set_attributes(self, params):
        self.uuid = params.get('uuid')
        self.email = params.get('email')
        self.password = params.get('password')
        self.first_name = params.get('firstName')
        self.last_name = params.get('lastName')
        self.phone_number = params.get('phone')
        self.locale = params.get('locale')
        self.timezone = params.get('timezone')
        self.platform = params.get('platform')
        self.role = params.get('role')
        self.archived = params.get('archived')
        self.custom_fields = params.get('customFields')
        self.address = params.get('address')


class Users(ListResource):
    """
    Class used for listing users from api
    """
    instance = User
    valid_filters = ['firstName']


class Item(InstanceResource):
    """
    Class for item instance resource
    """
    path = '/item/'
    valid_filters = [
        'sortedBy', 'name', 'partNumber', 'taxable', 'glAccount'
    ]

    def _set_attributes(self, params):
        self.uuid = params.get('uuid')
        self.name = params.get('name')
        self.part_number = params.get('partNumber')
        self.description = params.get('description')
        self.gl_account = params.get('glAccount')
        self.markup_type = params.get('markupType')
        self.markup_value = params.get('markupValue')
        self.taxable = params.get('taxable')
        self.track_on_van = params.get('trackOnVan')
        self.unit_cost = params.get('unitCost')
        self.unit_price = params.get('unitPrice')
        self.custom_fields = params.get('customFields')


class Items(ListResource):
    """
    Class used for listing item objects on api
    """
    instance = Item


class Customer(InstanceResource):
    path = '/customer/'
    valid_filters = ['sortedBy', 'firstName', 'lastName',
                      'email', 'phone', 'customer']


class Contact(InstanceResource):
    path = '/contact/'
    valid_filters = ['sortedBy', 'firstName', 'lastName',
                      'email', 'phone', 'customer']


class Location(InstanceResource):
    path = '/location/'
    valid_filters = ['sortedBy', 'name', 'streetName', 'locality',
                      'region', 'postcode', 'country', 'isBilling',
                      'type', 'customer']






from os.path import join
import os
from pytest import fixture
from vcr import VCR
from fawrapper.requestmaker import RequestMaker
from fawrapper import FieldawareAPIClient

api_path = '/'
host = 'api.fieldaware.net'
api_key = os.environ['FIELDAWARE_TOKEN']

@fixture
def api():
    return FieldawareAPIClient()

@fixture
def request_maker():
    return RequestMaker(host, api_key)


@fixture
def vcr_cassette_path(request, vcr_cassette_name):
    # USED BY VCRPY
    # Put all cassettes in tests/api_returns/{module}/{test}.json
    return join('tests/api_returns', request.module.__name__, vcr_cassette_name)


@fixture
def vcr(request, vcr_config):
    # USED BY VCRPY
    """The VCR instance"""
    kwargs = dict(
        path_transformer=VCR.ensure_suffix(".json"),
        serializer='json',
        filter_headers= [('Authorization', 'DUMMY')],
        #record_mode='once',
        record_mode='new_episodes',
        #record_mode='none',
    )
    marker = request.node.get_marker('vcr')
    record_mode = request.config.getoption('--vcr-record-mode')

    kwargs.update(vcr_config)
    if marker:
        kwargs.update(marker.kwargs)
    if record_mode:
        kwargs['record_mode'] = record_mode

    vcr = VCR(**kwargs)
    return vcr

@fixture
def recorded():
    return vcr



"""Test Jobs list from api"""
from pytest import mark, fixture, raises
from unittest import mock
from datetime import datetime
from unittest.mock import Mock

from fawrapper import FieldawareAPIClient
from fawrapper.utils import format_time_to_iso
from fawrapper.models import Jobs, Item
from fawrapper.exceptions import APIValidationError


@mark.vcr
def test_items_used_in_job(api):
    job_uuid = '427174cd4cd54f9e9cd35fe9a57b01c7'
    job = api.jobs.get(job_uuid)
    assert type(job.items_used) is list
    assert type(job.items_used[0]) is Item
    assert len(job.items_used) > 0
    assert job.items_used[0].name == 'Travel - Mileage'


@mark.vcr
def test_job_lead_change(api):
    job_uuid = 'aa0cf993283741ea97ec48924b17af60'
    job = api.jobs.get(job_uuid)

    filters = {'firstName': 'Johnny'}
    user = next(api.users.list(filters))
    assert user.first_name == 'Johnny'

    job.job_lead = user
    assert job.job_lead.first_name == 'Johnny'

    changed_job = api.jobs.get(job_uuid)
    assert changed_job.job_lead.first_name == 'Johnny'


@mark.vcr
def test_job_list_filters_completed(api):
    filters = {'state': 'completed'}
    job = next(api.jobs.list(filters))
    assert job.state == 'completed'


@mark.vcr
def test_job_list_filters_scheduled(api):
    filters = {'state': 'scheduled'}
    job = next(api.jobs.list(filters))
    assert job.state == 'scheduled'

@mark.vcr
def test_get_all_jobs(api):
    start = format_time_to_iso(2017, 1, 1)
    end = format_time_to_iso(2017, 9, 30)
    filters = {'start': start, 'end': end}
    job = next(api.jobs.list(filters))
    assert job.uuid is not None
    assert job.description is not None
    assert job.location is not None
    assert job.job_lead is not None

@mark.vcr
def test_get_all_jobs_no_filters(api):
    job = next(api.jobs.list())
    assert job.uuid is not None


def test_validate_empty_filters():
    with mock.patch('fawrapper.models.models.format_time_to_iso',
                    return_value='test_string'):
        returned_filters = Jobs()._validate_filters()
    assert type(returned_filters) is dict
    assert type(returned_filters['start']) is str
    assert returned_filters['start'] == 'test_string'


def test_validate_default_end_filter():
    with mock.patch('fawrapper.models.models.format_time_to_iso',
                    return_value='test_string'):
        returned_filters = Jobs()._validate_filters()
    assert type(returned_filters['end']) is str
    assert returned_filters['end'] == 'test_string'


def test_incorrect_filters_type():
    with raises(TypeError):
        Jobs()._validate_filters(filters=list('some_value'))


def test_invalid_filter_raises():
    invalid_filters = {'invalid_key': 'invalid_value'}
    with raises(Exception):
        Jobs()._validate_filters(invalid_filters)














"""Test the Tasks object on the api"""
from pytest import mark

@mark.vcr
def test_tasks_list(api):
    task = next(api.tasks.list())
    assert task.uuid == '0382b04b53e340feb3c054bad29d646f'
    assert task.gl_account is None
    assert task.name == 'Prover'
    assert task.markup_type == 'fixed'
    assert task.taxable == True
    assert task.markup_value == 0.0
    assert task.custom_fields == {}
    assert task.est_duration == None
    assert task.unit_price == 0.0
    assert task.unit_cost == 0.0
    assert task.description == 'Needed for calibration'




"""Tests for the utils module"""
from fawrapper.utils import format_time_to_iso


def test_format_time_to_iso_utc_string():

    datetime_string = format_time_to_iso(2017, 9, 30, 21, 37, 1)

    # TODO: specify timezone for testing on multiple zones
    #assert datetime_string == '2017-09-30T21:37:01-05:00'



"""Test the items resource on the api."""
from pytest import mark

from fawrapper import FieldawareAPIClient
from fawrapper.models import Item


@mark.vcr
def test_full_items_list():
    api = FieldawareAPIClient()
    items_list = api.items.list()
    item = next(items_list) # Only need to test first item
    assert isinstance(item, Item)
    assert item.name is not None
    assert item.part_number is not None






"""Test the Task object on the api"""

from fawrapper.models import TaskInJob, Task, ItemInTask, Item
from pytest import mark


@mark.vcr
def test_task_in_job_init(request_maker):
    job_uuid = '427174cd4cd54f9e9cd35fe9a57b01c7'
    task_uuid = '116efd341c1f4e42a3e3544c10c1c3ad'
    params = {'uuid': task_uuid, 'job_uuid': job_uuid}
    task = TaskInJob(requester=request_maker, **params)
    task.pull_details()
    assert task.uuid == task_uuid
    assert task.job_uuid == job_uuid
    assert task.attachments is not None
    assert task.unit_price == 0.0
    assert task.unit_cost == 0.0
    assert type(task.note) is str


@mark.vcr
def test_task_in_taskinjob_init(request_maker):
    job_uuid = '427174cd4cd54f9e9cd35fe9a57b01c7'
    task_uuid = '116efd341c1f4e42a3e3544c10c1c3ad'
    params = {'uuid': task_uuid, 'job_uuid': job_uuid}
    task = TaskInJob(requester=request_maker, **params)
    task.pull_details()
    assert type(task.task) is Task
    assert task.task.uuid == 'b99787ec09dd4c0ba84e26c41749c570'
    assert task.task.taxable == False
    assert task.task.name == 'Service and Repair'


@mark.vcr
def test_items_in_task_in_job(request_maker):
    job_uuid = '427174cd4cd54f9e9cd35fe9a57b01c7'
    task_uuid = '116efd341c1f4e42a3e3544c10c1c3ad'
    params = {'uuid': task_uuid, 'job_uuid': job_uuid}
    task_in_job = TaskInJob(requester=request_maker, **params)
    task_in_job.pull_details()
    assert type(task_in_job.items) is list
    assert len(task_in_job.items) > 0
    for each in task_in_job.items:
        assert isinstance(each, ItemInTask)
        assert each._item is not None
        assert each.uuid is not None

        # Defaults
        assert isinstance(each.item, Item)
        assert each.item.uuid is not None
        assert each.item.name is not None
        assert each.item.description is not None

        # Pulled from api
        assert each.item.part_number is not None
        assert type(each.item.unit_cost) is float
        assert type(each.item.unit_price) is float








from requests.exceptions import HTTPError


class FieldawareError(HTTPError):
    """Generic error while attempting to communicate with Fieldaware"""

class UnknownStatusCode(FieldawareError):
    """Status code has not been added to system"""

class APIValidationError(FieldawareError):
    """
        (422)
        The query parameters and/or body is not valid for this request.
        This error captures violation of structural assumptions made on the
        inputs, as captured by the different schemas that are included in the
        Reference section of this documentation.
    """

class ConflictError(FieldawareError):
    """
        (409)
        The server cannot satisfy an otherwise valid request because of a
        conflicting business logic rule (e.g. attempting to invoice a
        non-completed job). While an APIValidationError will never be
        acceptable if re-submitted, a request that gives rise to a
        ConflictError may be acceptable after other resource state changes
        (e.g. a job is completed by a field engineer).

    """

class AuthenticationError(FieldawareError):
    """"
        (401)
        The client failed to authenticate, typically because the API token was
        either invalid or not included in the request.
    """

class NotFoundError(FieldawareError):
    """"
        (404)
        A specified resource does not exist. This may be the top-level
        resource at which the request was directed, but it may also be one of
        the entities referenced in the payload. The error message should
        contain the exact reference that caused offence.
    """


class BadRequest(FieldawareError):
    """
        (400)
        The payload is syntactically malformed. For example, the inputs are
        not encoded as JSON, or the payload is not valid JSON.
    """

class APIKeyMissingError(Exception):
    pass


class UUIDMissingError(Exception):
    pass


class InvalidObjectType(Exception):
    pass



"""Test the item object on the api."""
from pytest import mark
from fawrapper import FieldawareAPIClient
from fawrapper.models import Item, ItemInTask

@mark.vcr
def test_get_item_with_uuid():
    api = FieldawareAPIClient()
    item = api.items.get('0121fca8d1314f4093f911e35d427581')
    assert isinstance(item, Item)
    assert item.name is not None
    assert item.part_number is not None


def test_item_in_task_in_job_init(request_maker):
        job_uuid = '427174cd4cd54f9e9cd35fe9a57b01c7'
        task_in_job_uuid = '116efd341c1f4e42a3e3544c10c1c3ad'
        params = {'task_in_job_uuid': task_in_job_uuid, 'job_uuid': job_uuid}
        item = ItemInTask(requester=request_maker, **params)




"""Base Classes for objects on the api"""


class Resource(object):
    """Base class for all resources on API"""
    path = None
    instance = None

    def __init__(self, requester=None):
        self.requester = requester

    def post(self, data=None):
        path = self.path
        if self.uuid:
            path = f'{path}{self.uuid}'
        return self.requester.post(path, data=data)

    def get(self, uuid=None):
        """Get an item based on uuid or filters"""
        path = self.instance.path
        if uuid:
            path = f'{path}{uuid}'
        data = self.requester.get(path)
        return self._parse_to_instance(data)

    def _parse_to_instance(self, data):
        return self.instance(self.requester, **data)


class ListResource(Resource):
    """Class for all resources that have multiple objects"""
    valid_filters = ['pageSize']

    def list(self, filters=None):
        """Get all items on api"""
        filters = self._validate_filters(filters)
        for page in self._get_all_pages(filters):
            for item in page:
                item = self._parse_to_instance(item)
                item.pull_details()
                yield item

    def _validate_filters(self, filters):
        if not filters:
            filters = dict()
        if type(filters) is not dict:
            raise TypeError('Filters should be type dict')

        self.valid_filters.extend(ListResource.valid_filters)
        for key in filters:
            if key not in self.valid_filters:
                raise KeyError(f'Filter ({key}) not a valid filter')
        return filters

    def _get_all_pages(self, filters=None):
        params = {'page': 0, 'pageSize': 100}
        if filters:
            params.update(filters)
        items_returned = 1
        while items_returned:
            page = self.requester.get(
                self.instance.path, params=params
            )
            items_returned = len(page['items'])
            params['page'] += 1
            if items_returned:
                yield page['items']


class InstanceResource(Resource):
    """Class for individual resources"""
    uuid = None

    def __init__(self, requester=None, **params):
        super().__init__(requester)
        self._set_attributes(params)
        self._raw_params = params

    def pull_details(self):
        """
        Pull details from api and update attributes with data
        """
        path = f'{self.path}{self.uuid}'
        data = self.requester.get(path)
        self._set_attributes(data)


    def _set_attributes(self, params):
        """
        Update all attributes of an object
        from a list of properties returned from api
        """
        raise NotImplemented



from pytest import mark

from fawrapper import FieldawareAPIClient


@mark.vcr
def test_basic_users_list():
    api = FieldawareAPIClient()
    users = [user for user in api.users.list()]
    for user in users:
        assert hasattr(user, 'uuid')
        assert hasattr(user, 'email')
        assert hasattr(user, 'first_name')
        assert hasattr(user, 'last_name')


@mark.vcr
def test_users_list_with_filters():
    api = FieldawareAPIClient()
    filters = {'firstName': 'faadmin'}
    users = api.users.list(filters)
    for each in users:
        assert each.first_name == 'faadmin'


@mark.vcr
def test_get_user_with_uuid():
    api = FieldawareAPIClient()
    uuid = '145a5454958d45c0956f0e36126363ac'
    user = api.users.get(uuid)
    assert user.email == 'faadmin@aphelpspetroleum.com'


def test_user_parse_from_dict():
    data = {
        'uuid': 'user_uuid',
        'email': 'user@email.com',
        'firstName': 'user_name'
    }
    api = FieldawareAPIClient('some_key')
    user = api.users._parse_to_instance(data)
    assert user.uuid == 'user_uuid'
    assert user.email == 'user@email.com'
    assert user.first_name == 'user_name'









from urllib.parse import urlencode
import json

import requests

from fawrapper.exceptions import (
    APIValidationError, AuthenticationError,
    BadRequest, ConflictError, NotFoundError, UnknownStatusCode)


def build_url(hostname, path, params=None, protocol='https://'):
    url =  f'{protocol}{hostname}{path}'
    if params:
        url += '?' + urlencode(params)
    return url
        

class RequestMaker:
    def __init__(self, host, api_key):
        self.host = host
        self.api_key = api_key

        self.headers = {"Authorization": f"Token {self.api_key}",
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                        "Host": self.host}

    def _request(self, method, path, data=None, params=None):
        url = build_url(self.host, path, params)
        #import pdb; pdb.set_trace()
        response = requests.request(
            headers=self.headers,
            method=method,
            url=url,
            data=data or {},
        )
        self._check_status_code(response)
        try:
            return response.json()
        except json.JSONDecodeError:
            return response.text

    def get(self, uri, params=None):
        """Retrieve an object(s) from api"""
        return self._request('get', uri, params=params)

    def post(self, uri, data=None):
        """Create a new object"""
        if type(data) is dict:
            data = json.dumps(data)
            self._request('post', uri, data)

    def put(self, uri, data=None):
        """Update an existing object"""
        self._request('put', uri, data)

    def delete(self, uri):
        """Delete an existing object"""
        self._request('delete', uri)

    def _check_status_code(self, response):
        error_codes = {
            422: APIValidationError,
            401: AuthenticationError,
            400: BadRequest,
            409: ConflictError,
            404: NotFoundError,
        }
        if response.status_code in error_codes:
            raise error_codes[response.status_code](
                response.text , response.url)

        success_codes = {
            200: 'OK',
            204: 'No Content',
            201: '?',
        }
        if response.status_code not in success_codes:
            raise UnknownStatusCode(
                f'{response.status_code}: {response.text}, {response.url}')



"""Utilities for data management and interacting with the api"""
from datetime import datetime, timezone
import os


def persist_token(token, file='token_db'):
    with open(file, 'w') as f:
        f.write(token)


def read_token(file='token_db'):
    with open(file, 'r') as f:
        return f.read()


def format_time_to_iso(year, month, day,
                       hour=0, minute=0, second=0):
    """
    Convert local time to UTC time with TZ offset,
    in proper UTC format required by the api
    """
    datetime_obj = datetime(year, month, day, hour, minute, second)
    return datetime_obj.astimezone().isoformat()



"""Test Job objects on the api"""
from pytest import mark
from fawrapper.models import Jobs, User, TaskInJob
from fawrapper import FieldawareAPIClient


@mark.vcr
def test_job_lead_attribute_is_user_instance():
    api = FieldawareAPIClient()
    job = api.jobs.get('7b6c125c296449ed9bf4763aaf0dbaa0')
    assert isinstance(job.job_lead, User)
    assert job.job_lead.uuid == 'a2a028db0a064ef0b95d57becd7c1e0c'
    assert job.job_lead.first_name == 'Office'


@mark.vcr
def test_tasks_attribute_is_list_of_instances():
    api = FieldawareAPIClient()
    job = api.jobs.get('7b6c125c296449ed9bf4763aaf0dbaa0')
    assert isinstance(job.tasks, list)
    for each in job.tasks:
        assert isinstance(each, TaskInJob)




from fawrapper.models import Users, Items, Jobs, Tasks, Changes
from fawrapper.requestmaker import RequestMaker
import os


class FieldawareAPIClient:
    def __init__(self, api_key=None, host="api.fieldaware.net"):
        self.api_key = api_key or os.environ['FIELDAWARE_TOKEN']
        self.host = host

        self.raw_request = RequestMaker(self.host, self.api_key)
        self.requester = RequestMaker(self.host, self.api_key)
        self._init_resources()


    def _init_resources(self):
        self.users = Users(self.raw_request)
        self.items = Items(self.raw_request)
        self.jobs = Jobs(self.raw_request)
        self.tasks = Tasks(self.raw_request)
        self.changes = Changes(self.raw_request)





from setuptools import setup, find_packages

with open('README.rst', 'r') as f:
    readme = f.read()

setup(
    name="fawrapper",
    version="0.1.0a4",
    packages=find_packages(),
    description='A wrapper for the Fieldaware API',
    long_description=readme,
    author='Johnny Kirchman',
    author_email='Jkirch86@gmail.com',
    license='MIT',
    url='https://github.com/J-Kirch/fawrapper',
    python_requires='>3.6',
    install_requires=['requests'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'vcrpy', 'pytest-vcr'],
    keywords=[],
    classifiers=[],
)



"""Test changes endpoint on api"""
from fawrapper.utils import persist_token, read_token


def test_token_save_and_read_on_disk(tmpdir):
    token = '1234567'
    file = tmpdir.mkdir('fawrapper/').join('tmp_token_db')
    persist_token(token, file=file)
    assert read_token(file=file) == token


def test_get_token(api):
    api.changes.list()





from unittest.mock import patch
import json

import pytest

from fawrapper.exceptions import (
    APIValidationError, AuthenticationError,
    BadRequest, ConflictError, NotFoundError, UnknownStatusCode)
from fawrapper.requestmaker import build_url
from tests.conftest import api_key, host
from .tools import MockResponse


def test_minimum_init(request_maker):
    assert request_maker.host == host
    assert request_maker.api_key == api_key

    assert request_maker.headers == {
        'Authorization': f'Token {api_key}',
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        'Host': request_maker.host,
    }


def test_build_url():
    url = build_url(protocol='https://', hostname='api.net', path='/')
    assert url == 'https://api.net/'

    url = build_url('api.net', '/path/')
    assert url == 'https://api.net/path/'

    url = build_url('api.net', '/path/', params={'key': 'value'})
    assert url == 'https://api.net/path/?key=value'


@patch('fawrapper.requestmaker.requests.request')
def test_requests_get_is_called(requests_request, request_maker):
    requests_request.return_value = MockResponse(200, "test")
    request_maker.get('/something')
    requests_request.assert_called_with(
        method='get',
        url=f'https://{host}/something',
        data={},
        headers=request_maker.headers,
    )


@patch('fawrapper.requestmaker.requests.request')
def test_requests_get_is_called_with_params(requests_request, request_maker):
    requests_request.return_value = MockResponse(200, '""')
    params = {'key': 'value'}
    request_maker.get('/something/', params)
    requests_request.assert_called_with(
        method='get',
        url=f'https://{host}/something/?key=value',
        data={},
        headers=request_maker.headers,
    )


@patch('fawrapper.requestmaker.requests.request')
def test_requests_post_is_called(requests_request, request_maker):
    requests_request.return_value = MockResponse(200, '""')
    request_maker.post('/something/', {'key': 'value'})
    requests_request.assert_called_with(
        method='post',
        url=f'https://{request_maker.host}/something/',
        data=json.dumps({"key": "value"}),
        headers=request_maker.headers,
    )


@patch('fawrapper.requestmaker.requests.request')
def test_requests_put_is_called(requests_request, request_maker):
    requests_request.return_value = MockResponse(200, '""')
    request_maker.put('/something/', {'key': 'value'})
    requests_request.assert_called_with(
        method='put',
        url=f'https://{request_maker.host}/something/',
        data={"key": "value"},
        headers=request_maker.headers,
    )


@patch('fawrapper.requestmaker.requests.request')
def test_requests_delete_is_called(requests_request, request_maker):
    requests_request.return_value = MockResponse(200, '""')
    request_maker.delete('/something/')
    requests_request.assert_called_with(
        method='delete',
        url=f'https://{request_maker.host}/something/',
        data={},
        headers=request_maker.headers,
    )


@pytest.mark.parametrize('status_code,exception', [
    (422, APIValidationError),
    (409, ConflictError),
    (401, AuthenticationError),
    (404, NotFoundError),
    (400, BadRequest)])
def test_error_status_codes(status_code, exception, request_maker):
    with pytest.raises(exception) as error:
        response = MockResponse(status_code, 'error message')
        request_maker._check_status_code(response)
    assert error.value.args[0] == 'error message'


@pytest.mark.parametrize('status_code', [200, 201, 204])
def test_200_status_code_passes(status_code, request_maker):
    response = MockResponse(status_code, '""')
    request_maker._check_status_code(response)


def test_unknown_status_code_raises_exception(request_maker):
    with pytest.raises(UnknownStatusCode) as error:
        response = MockResponse(888, 'unknown error message')
        request_maker._check_status_code(response)
    assert error.value.args[0] == '888: unknown error message, mock_url'


